import cv2

from yarok.core.config_block import ConfigBlock


class Cv2WaitKey:

    def __init__(self, config: ConfigBlock):
        self.config = config
        self.wait_ms = self.config['wait_ms'] if 'wait_ms' in self.config else 1
        self.skip_s = (self.config['skip_ms'] if 'skip_ms' in self.config else 100) / 1000
        self.next_probe_ts = time.time()

    def step(self):
        if time.time() > self.next_probe_ts:
            cv2.waitKey(self.config['ms'] or 1)
            key = cv2.waitKey(self.wait_ms)
            self.next_probe_ts = time.time() * self.skip_s



from xml.etree.ElementTree import tostring

from yarok.core.components_manager import ComponentsManager
from yarok.core.config_block import ConfigBlock
from yarok.core.platform import Platform
from yarok.core.injector import Injector

import inspect


class PlatformHW(Platform):
    """
        Is the Platform for the Real World

    """

    def __init__(self, manager: ComponentsManager, config: ConfigBlock, behaviour):
        super(PlatformHW, self).__init__(manager, config, behaviour)
        self.interfaces = {
            id: self.init_interface(id, c)
            for id, c in self.manager.components.items()
            if 'interface_hw' in self.manager.config(id)}

        self.init_components(self.interfaces)

        while not self.interfaces_ready():
            self.step()

    def init_interface(self, n, c):
        runtime_conf = ConfigBlock(
            self.config['interfaces'][c['name_path']] if
            'interfaces' in self.config and c['name_path'] in self.config['interfaces']
            else {})

        interface_cls = self.manager.config(n)['interface_hw']

        # backwards compatible:
        if not hasattr(self.manager.config(n)['interface_hw'], '__data__'):
            print('[deprecated] interfaces should be annotated with the @interface decorator')
            return interface_cls(interfaces_mjc[n])

        class_members = {t[0]: t[1] for t in inspect.getmembers(interface_cls)}
        init_members = {t[0]: t[1] for t in inspect.getmembers(class_members['__init__'])}
        init_annotations = init_members['__annotations__'] if '__annotations__' in init_members else {}

        interface_defaults = interface_cls.__data__['defaults']
        config = ConfigBlock(runtime_conf)
        config.defaults(interface_defaults)

        providers = [
            {'match_type': ConfigBlock, 'value': config},
        ]

        injector = Injector(self, providers=providers)

        return interface_cls(**injector.get_all(init_annotations))

    def interfaces_ready(self):
        return not (sum(
            [not self.interfaces[i].is_ready() for i in self.interfaces if hasattr(self.interfaces[i], 'step')]) > 0)

    def step(self):
        """
            Steps through all the interfaces
        """
        _ = {self.interfaces[i].step() for i in self.interfaces if hasattr(self.interfaces[i], 'step')}


class Injector:

    def __init__(self, platform, component_ref=None, config=None, providers=None):
        self.providers = providers if providers is not None else []
        self.platform = platform
        self.component_ref = component_ref
        self.config = config if config is not None else self.component_ref[
            'config'] if self.component_ref is not None else None

    def get(self, name_or_cls=None, cls=None):
        name = name_or_cls if isinstance(name_or_cls, str) else None
        injectable_cls = name_or_cls if not isinstance(name_or_cls, str) else cls

        if injectable_cls is None and self.component_ref and self.component_ref['name_path'] == '/' and name == 'world':
            return self.component_ref['instance']
        elif injectable_cls is None and self.component_ref:
            return self.platform.manager.get_by_name(self.component_ref, name)['instance']
        else:
            if isinstance(self.platform, injectable_cls):
                return self.platform
            elif isinstance(self, injectable_cls):
                return self
            elif isinstance(self.config, injectable_cls):
                return self.config
            elif name is not None and self.component_ref:
                return self.platform.manager.get_by_name(self.component_ref, name)['instance']

        for provider in self.providers:
            if 'match_type' in provider and provider['match_type'] == injectable_cls:
                return provider['value']

        return None

    def get_all(self, args):
        if type(args) is dict:
            return {a: self.get(a, args[a]) for a in args}



from yarok import ConfigBlock
from yarok.core.components_manager import component
from yarok.platforms.mjc.interface import InterfaceMJC

import serial


class AnetA30InterfaceMJC:

    def __init__(self, interface: InterfaceMJC):
        self.interface = interface
        self.gear = 100
        self.MAX_X = 320
        self.MAX_Y = 320
        self.MAX_Z = 420

        # includes gear ratio and millimeters to meters conversion.
        self.mm2ctrl = self.gear / 1000.0

    # moves the printer to the initial position
    # implemented to match the same api as the real printer.
    # it isn't necessary for simulation
    def home(self):
        pass

    def move(self, p):
        x, y, z = [
            max(0, min(p[0], self.MAX_X)),
            max(0, min(p[1], self.MAX_Y)),
            max(0, min(p[2], self.MAX_Z))
        ]

        self.interface.set_ctrl('ax', x * self.mm2ctrl)
        self.interface.set_ctrl('ay', y * self.mm2ctrl)
        self.interface.set_ctrl('az', z * self.mm2ctrl)

    def is_at(self, position):
        sensor_data = [d / self.mm2ctrl for d in self.interface.sensordata()]
        # print(sensor_data, sum([abs(position[i] - sensor_data[i]) for i in range(3)]))
        return sum([abs(position[i] - sensor_data[i]) for i in range(3)]) < 0.25  # 0.25 mm

    def step(self):
        pass


class AnetA30InterfaceHW:

    def __init__(self, config: ConfigBlock):
        config.defaults({
            'serial_path': '/dev/ttyUSB0'
        })
        self.ser_con = serial.Serial(config['serial_path'], 115200, timeout=1)

        self.MAX_X = 320
        self.MAX_Y = 320
        self.MAX_Z = 420
        self.initialized = False
        self.position = None

        self.uninitialized_empty_counter = 0  # used to keep track of the initialization
        self.wait_until_initialize = 5
        self.initialized = False

        self.executing_cmd = False
        self.last_exec_cmd = ''

    def gcode_parse_pos(self, res):
        arr = res.split(' ')
        pos_d = {a.split(':')[0]: float(a.split(':')[1]) for a in arr}
        return pos_d['X'], pos_d['Y'], pos_d['Z']

    def send_cmd(self, cmd, args=None):
        if not self.initialized:
            return False
        full_cmd = cmd + " " + args if args is not None else cmd
        self.ser_con.write(str.encode(full_cmd + ' \n'))
        self.ser_con.flush()

        self.executing_cmd = True
        self.last_exec_cmd = cmd

        return True

    def home(self):
        # implemented by the interface
        pass

    def is_at(self, position):
        if not self.is_moving() and \
                (position is not None) and (self.position is not None) and \
                sum([abs(position[i] - self.position[i]) for i in range(3)]) < 0.25:  # 0.25 mm
            return True
        return False

    def is_moving(self):
        return (self.last_exec_cmd == 'G0' or self.executing_cmd == 'G28') \
               and self.executing_cmd

    def move(self, position: [float, float, float]):
        x, y, z = round(position[0], 2), round(position[1], 2), round(position[2], 2)
        x = max(0, min(x, self.MAX_X))
        y = max(0, min(y, self.MAX_Y))
        z = max(0, min(z, self.MAX_Z))
        if self.send_cmd("G0", "X%.2f Y%.2f Z%.2f" % (x, y, z)):
            self.position = None

    def is_ready(self):
        if not self.initialized:
            return False
        elif not self.executing_cmd:
            if self.last_exec_cmd != 'G28':
                # home the printer
                self.send_cmd('G28')
            else:
                return True
        return False

    def step(self):
        ln = self.ser_con.readline().decode().rstrip()
        if not self.initialized:
            self.uninitialized_empty_counter = self.uninitialized_empty_counter + 1 if ln == '' else 0
            if self.uninitialized_empty_counter >= self.wait_until_initialize:
                self.initialized = True
        else:
            if self.executing_cmd:
                if ln == 'ok':
                    self.executing_cmd = False
                elif self.last_exec_cmd == 'M114' and ln != '':
                    self.position = self.gcode_parse_pos(ln)
            else:
                if self.last_exec_cmd != 'M114':
                    self.send_cmd('M114')


@component(
    tag="anet_a30",
    interface_mjc=AnetA30InterfaceMJC,
    interface_hw=AnetA30InterfaceHW,
    # language=xml
    template="""
    <mujoco>
    <asset>
        <mesh name="fdm_printer_bottom_frame" file="meshes/bottom_frame.stl" scale="0.0011 0.0011 0.0011"/>
        <mesh name="fdm_printer_side1_frame" file="meshes/side1_frame.stl" scale="0.0011 0.0011 0.0011"/>
        <mesh name="fdm_printer_side2_frame" file="meshes/side2_frame.stl" scale="0.0011 0.0011 0.0011"/>


        <mesh name="fdm_printer_bed" file="meshes/bed2.stl" scale="0.0011 0.0011 0.0011"/>
        <mesh name="fdm_printer_x_axis" file="meshes/x_axis.stl" scale="0.0011 0.0011 0.0011"/>

        <material name="black_metal" rgba=".2 .2 .2 1" specular="0.95"/>

    </asset>
    <worldbody>
        <body>

            <!-- External frame -->
            <body pos="0 0 0.09">
                <geom type="mesh"
                      mesh="fdm_printer_bottom_frame"
                      material="black_metal"/>
                <geom type="mesh"
                      mesh="fdm_printer_side1_frame"
                      material="black_metal"/>
                <geom type="mesh"
                      mesh="fdm_printer_side2_frame"
                      material="black_metal"/>
                <geom type="mesh"
                      mesh="fdm_printer_x_axis"
                      material="black_metal"/>
            </body>

            <!-- Z-AXIS: horizontal metal bar -->
            <body name="zaxis_body" pos="0 -0.025 0.0634">

                <joint name="zaxis"
                       type="slide"
                       armature="10"
                       frictionloss="100"
                       axis="0 0 1"/>

                <geom type="mesh"
                      mesh="fdm_printer_x_axis"
                      material="black_metal"
                      fitscale="0.001"
                      pos="0 0 -0.498"/>

                <!-- x-axis -->
                <body name="xaxis_body">
                    <joint name="xaxis"
                           type="slide"
                           frictionloss="100"
                           armature="10"
                           axis="1 0 0"/>

                    <!-- printer head-->
                    <body pos="-0.002 0 0.05" name="printer_head">
                        <geom type="box"
                              pos="0.001 0.01 0.005"
                              size=".04 .002 .035"
                              material="black_metal"/>
                        <!-- <geom type="box"
                              pos="-0.01 -0.014 -0.015"
                              size=".025 .018 .0125"
                              material="black_plastic"/> -->
                              
                              
                    </body>
                </body>
            </body>

            <body name="printer_bed" pos="0 0 0.08">
                <joint name="yaxis"
                       type="slide"
                       armature="1"
                       frictionloss="100"
                       axis="0 -1 0"/>

                <!-- area/size of the real printer bed -->
                <!-- <geom type="box"
                      name="fake_bed"
                      size="0.165 0.165 0.001"
                      pos="0.164 .164 -0.01"
                      rgba="0.1 .1 0 .5"/> -->

                <geom type="mesh"
                      mesh="fdm_printer_bed"
                      material="black_metal"
                      friction="0.4 0.4 0.8"/>


            </body>
        </body>
    </worldbody>
    <actuator>
        <position name="ax" gear="100" joint="xaxis" forcelimited="true" forcerange="-1.02 1.02"/>
        <position name="ay" gear="100" joint="yaxis" forcelimited="true" forcerange="-0.2 0.2"/>
        <position name="az" gear="100" joint="zaxis" forcelimited="true" forcerange="-0.55 0.55"/>
    </actuator>
    <sensor>
        <actuatorpos name="x" actuator="ax"/>
        <actuatorpos name="y" actuator="ay"/>
        <actuatorpos name="z" actuator="az"/>
    </sensor>
</mujoco>
    """
)
class AnetA30:

    def __init__(self):
        pass

    def home(self):
        # implemented by the interface
        pass

    def is_at(self, position):
        # implemented by the interface
        pass

    def move(self, position: [float, float, float]):
        # in millimeters (mm)
        # implemented by the interface
        pass



def behaviour(**kwargs):
    def _(c):
        if not hasattr(c, '__data__'):
            defaults = kwargs['defaults'] if 'defaults' in kwargs else {}

            setattr(c, '__data__', {
                'defaults': defaults,  # behaviour config defaults
            })
        return c

    return _



from yarok import Platform, PlatformMJC, PlatformHW, component, ConfigBlock, Injector

from yarok.comm.worlds.empty_world import EmptyWorld
from yarok.comm.components.robotiq_2f85.robotiq_2f85 import Robotiq2f85
from yarok.comm.components.ur5.ur5 import UR5
from yarok.comm.components.geltip.geltip import GelTip
from yarok.comm.components.ur5e.ur5e import UR5e
from yarok.comm.components.cam.cam import Cam

from yarok.comm.plugins.cv2_inspector import Cv2Inspector
from yarok.comm.plugins.cv2_waitkey import Cv2WaitKey

from math import pi

import cv2

@component(
    extends=EmptyWorld,
    components=[
        UR5e,
        Robotiq2f85,
        GelTip,
        Cam
    ],
    # language=xml
    template="""
        <mujoco>
            <default>
                <default class='marker'>
                    <geom type="box" 
                        size="0.02 0.02 0.001" 
                        conaffinity='32' 
                        contype='32'/>
                </default>
            </default>
            <worldbody>
                
                <body pos="0.5 0.5 0.4"> 
                    <composite type="grid" count="30 1 1" spacing="0.009" offset="0 0 1">
                        <!-- <skin material="matplane" inflate="0.001" subgrid="3" texcoord="true"/> -->
                        <joint kind="main" damping="0.001" frictionloss='0.0001'/>
                        <tendon kind="main" width="0.019" rgba=".8 .2 .1 1"/>
                        <geom size=".018" rgba=".8 .2 .1 1" mass='0.0001' 
                            friction='0.1 0.005 0.0001'/><!--
                            -->
                        <!-- <pin coord="1"/> -->
                        <!-- <pin coord="39"/> -->
                    </composite>
                </body>
                
                <body>
                    <geom type="box" size="1.0 1.0 0.05" pos="0 0 0.45"/>
                    <geom class="marker" pos="0.7 0.7 0.5" rgba='255 0 0 1'/>
                    <geom class="marker" pos="0.1 0.7 0.5" rgba='0 255 0 1'/>
                    <geom class="marker" pos="0.1 0.3 0.5" rgba='0 0 255 1'/>
                    <geom class="marker" pos="0.7 0.3 0.5" rgba='255 0 255 1'/>
                   <!-- <geom type="box" size=".025 .025 1.0" pos="1.0 1.0 1.0"/> -->
                </body>
                
                <body  pos='0.9 0.9 1' euler='1.6 2.35619449019 0'>
                   <body euler='2.35619449019 0 3.14159265359'>
                     <cam name='vis_cam'/>
                   </body>
                </body>
                
                <body pos='0 0 0.5' xyaxes='0 1 0 -1 0 0'>
                    <ur5e name="arm">
                       <robotiq-2f85 name="gripper" parent="ee_link"> 
                             <body pos="0.02 -0.017 0.053" xyaxes="0 -1 0 1 0 0" parent="right_tip">
                                <geltip name="right_geltip" parent="left_tip"/>
                            </body>
                           <body pos="-0.02 -0.017 0.053" xyaxes="0 1 0 -1 0 0" parent="left_tip">
                                <geltip name="left_geltip" parent="right_tip"/>
                            </body>
                        </robotiq-2f85> 
                    </ur5e> 
                </body>
            </worldbody>        
        </mujoco>
    """
)
class GraspRopeWorld:
    pass


class GraspRoleBehaviour:

    def __init__(self, pl: Platform, injector: Injector):
        self.arm: UR5e = injector.get('arm')
        self.gripper: Robotiq2f85 = injector.get('gripper')
        self.pl = pl

    def on_update(self):
        # self.pl.wait_seconds(60)
        # print('------------------------------------------------------------------------------------------->')

        self.gripper.close(0.5)

        self.pl.wait(
            self.arm.move_xyz(xyz=[0.45, 0.5, 0.2], xyz_angles=[3.11, -1.6e-7, -pi / 2])
        )

        self.pl.wait(
            self.arm.move_xyz(xyz=[0.45, 0.5, 0.17], xyz_angles=[3.11, -1.6e-7, -pi / 2])
        )
        self.pl.wait_seconds(2)
        self.pl.wait(
            self.gripper.close(0.75)
        )
        self.pl.wait_seconds(10)
        print('closed gripper')

        self.pl.wait(
            self.arm.move_xyz(xyz=[0.45, 0.5, 0.4], xyz_angles=[3.11, -1.6e-7, -pi / 2])
        )

        self.pl.wait(
            self.arm.move_xyz(xyz=[0.45, 0.2, 0.4], xyz_angles=[3.11, -1.6e-7, -pi / 2])
        )
        self.pl.wait(
            self.arm.move_xyz(xyz=[0.45, 0.2, 0.25], xyz_angles=[3.11, -1.6e-7, -pi / 2])
        )
        self.pl.wait(
            self.gripper.close(0.5)
        )


        print('ended.')
        self.pl.wait_seconds(100)


conf = {
    'world': GraspRopeWorld,
    'behaviour': GraspRoleBehaviour,
    'defaults': {
        'environment': 'sim',
        'behaviour': {
        },
        'components': {
            '/gripper': {
                'left_tip': False,
                'right_tip': False,
            },
            '/right_geltip': {
                'label_color': '255 0 0'
            },
            '/left_geltip': {
                'label_color': '0 255 0'
            },
        },
    },
    'environments': {
        'sim': {
            'platform': {
                'class': PlatformMJC,
                'width': 800,
                'height': 600
            },
            'plugins': [
                (Cv2Inspector, {})
            ]
        }
    },
}

if __name__ == '__main__':
    Platform.create(conf).run()



from yarok import component


@component(
    tag='bucket-particles',
    defaults={
        'pos': [-0.135, -0.52, 0.165]
    },
    # language=xml
    template="""
        <mujoco>
            <asset>
                <material name='glass' rgba='1 1 1 0.1'/>
                <material name='red' rgba='.8 .2 .1 1'/>
                <material name='green' rgba='.2 .8 .1 1'/>
            </asset>
            <default>
                <default class='cup_side'>
                    <geom type="box" 
                        material='glass' 
                        size="0.0025 0.0220 0.0500" 
                        pos="0.05 0.0 0.0500" 
                        euler='0 0.05 0'/>
                </default>
                <default class='cup_bottom'>
                    <geom type="box" material='glass' size="0.0220 0.0500 0.0025"/>
                </default>
            </default>
            <worldbody>
                    <body pos='${pos[0]} ${pos[1]} ${pos[2]+0.05}'>
                        <body zaxis='1 1 1'>
                            <composite type="particle" count="4 4 4" spacing="0.0125">
                                <geom size="0.008" material='red'/>
                            </composite>
                        </body>
                    </body>   
                    
                    <body pos='${pos[0]} ${pos[1]} ${pos[2]+2*0.05}'>
                 <!--   zaxis='1 1 1' -->
                        <body >
                            <composite type="particle" count="4 4 4" spacing="0.02">
                                <geom size="0.008" material='green' mass='0.0001'/>
                            </composite>
                        </body>
                    </body>               
                 
    
                 <body pos='-0.135 -0.52 0.165'>
                        <freejoint/> 
                        <body euler='0 0 0.0'>
                            <geom class='cup_side'/>
                        </body>
                        <body euler='0 0 0.78539816339'>
                            <geom class='cup_side'/>
                        </body> 
                        <body euler='0 0 1.57079632679'>
                            <geom class='cup_side'/>
                        </body> 
                        <body euler='0 0 2.35619449019'>
                            <geom class='cup_side'/>
                        </body> 
                        <body euler='0 0 3.14159265359'>
                            <geom class='cup_side'/>
                        </body> 
                        <body euler='0 0 0.78539816339'>
                            <geom class='cup_side'/>
                        </body> 
                        <body euler='0 0 3.92699081699'>
                            <geom class='cup_side'/>
                        </body> 
                        <body euler='0 0 4.71238898038'>
                            <geom class='cup_side'/>
                        </body> 
                        <body euler='0 0 5.49778714378'>
                            <geom class='cup_side'/>
                        </body> 
                        
                        <body>
                            <geom class='cup_bottom' euler='0 0 -0.78539816339'/>
                            <geom class='cup_bottom' euler='0 0 0'/>
                            <geom class='cup_bottom' euler='0 0 0.78539816339'/>
                            <geom class='cup_bottom' euler='0 0 1.57079632679'/>
                        </body> 
                    </body>
            </worldbody>
        </mujoco>
    """
)
class BucketParticles:

    def __init__(self):
        pass



from yarok.core.components_manager import ComponentsManager
from yarok.core.config_block import ConfigBlock
from yarok.core.component import component
from yarok.core.interface import interface
from yarok.core.behaviour import behaviour
from yarok.core.platform import Platform
from yarok.core.injector import Injector

from yarok.platforms.mjc.platform import PlatformMJC
from yarok.platforms.hw.platform import PlatformHW

__PLATFORM__ = None
__VERSION__ = '0.0.35'


def wait(fn=None, cb=None):
    global __PLATFORM__
    print('[deprecated] wait_once was replaced by platform.wait')
    __PLATFORM__.wait(fn, cb)


def wait_seconds(seconds, cb=None):
    global __PLATFORM__
    print('[deprecated] wait_seconds was replaced by platform.wait_seconds')
    __PLATFORM__.wait_seconds(seconds, cb)


def wait_forever(cb=None):
    global __PLATFORM__
    print('[deprecated] wait_forever was replaced by platform.wait_forever')
    __PLATFORM__.wait_forever(cb)


def wait_once(cb=None):
    global __PLATFORM__
    print('[deprecated] wait_once was replaced by platform.wait_once')
    __PLATFORM__.wait_once(cb)


def load(config: {}):
    global __PLATFORM__
    print('[deprecated] yarok.load was replaced by Platform.create')
    __PLATFORM__ = Platform.create(config)
    return __PLATFORM__


def run(config={}):
    global __PLATFORM__
    print('[deprecated] run was replaced by Platform.create().run()')
    __PLATFORM__ = Platform.create(config)
    __PLATFORM__.run()
